@@ -67,46 +67,393 @@ class LearningAccessor(BaseAccessor):
 	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
 	"""
 
-	def _update_target(self, target, predictions):
-		'''
-		'''
-		if self.additive_learning:
-			if self.problem_type == 'regression':
-				return target-predictions if self.regression_error_type == 'additive' else target/predictions
-
-			if self.problem_type == 'classification':
-				return np.logical_not(target == predictions).astype(int)
-
-		else:
-			return target
-
-	def _compound_predictions(self, previous_predictions, new_predictions):
-		'''
-		'''
-		if self.additive_learning:
-			if self.problem_type == 'regression':
-				return previous_predictions+new_predictions if self.regression_error_type == 'additive' else \
-					previous_predictions*new_predictions
-
-			if self.problem_type == 'classification':
-				return np.abs(previous_predictions-new_predictions)
-
-		else:
-			return new_predictions
 
 
-	def _non_additive_fit(self):
-		pass
+	def _non_additive_fit(self, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
+			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=None, anonymize=False, \
+			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=1, \
+			regression_baseline='mean', regression_error_type='additive', return_scores=False, start_n_features_perf_frac=0.9):
+		# A base learner here does not fix mistakes made by another.
+		obj = self._obj
+		assert inspect.isfunction(learner_func), 'learner_func should be a class'
+		assert target_column in obj.columns, 'The target column should be a valid column'
+		if problem_type is None:
+			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'
+		assert problem_type in ('classification', 'regression')
+		self.problem_type = problem_type
+		self.additive_learning = False
+		assert regression_error_type in ('additive', 'multiplicative')
+		self.regression_error_type = regression_error_type
+
+		for col in obj.columns:
+			assert not self.is_categorical(col), 'All columns should be numeric'
+
+		x_columns = [_ for _ in obj.columns if _ != target_column]
+		if self.problem_type == 'classification':
+			labels = set(list(obj[target_column].values.astype(int)))
+			binary_labels = {0, 1}
+			assert labels.issubset(binary_labels), 'Classification labels should either be 0 or 1'
+
+		if benchmark_feature:
+			assert benchmark_feature in obj.columns, 'The benchmark feature should be a valid column'
+		self.benchmark_feature = benchmark_feature
+		if score == 'auto':
+			score = 'r2_score' if problem_type == 'regression' else 'accuracy_score'
+		score_func = eval(score)
+		self.val_scores = []
+
+		if getattr(self, 'models', None) is None or force_redo:
+			# 0. Train/Validation split
+			self.target_column = target_column
+			if return_scores:
+				# Reserve 1-train_frac for testing, and train_frac for training and validation
+				self.test_df = obj.sample(frac=1.-train_frac, random_state=random_state)
+				self.train_val_df = obj.drop(self.test_df.index)
+			else:
+				self.train_val_df = obj
+			# Reserve train_frac*train_frac for training [...]
+			self.train_df = self.train_val_df.sample(frac=train_frac, random_state=random_state)
+			# [...] and train_frac*(1-train_frac) for validation.
+			self.val_df = self.train_val_df.drop(self.train_df.index)
+
+			if missing_value_imputation:
+				# Basic missing value imputation
+				self.train_df.fillna(self.train_df.median(), inplace=True)
+				self.val_df.fillna(self.train_df.median(), inplace=True)
+				if return_scores:
+					self.test_df.fillna(self.train_df.median(), inplace=True)
+
+			# 1. Model-free variable selection
+			vs_accessor = PreLearningAccessor(obj)
+			self.variable_selection_results = vs_accessor.variable_selection(self.target_column, problem_type=self.problem_type, \
+				snr=snr, anonymize=anonymize)
+
+			self.variables = [_ for _ in self.variable_selection_results['Variable'].values if _.lower() != 'no variable']
+			n_variables = len(self.variables)
+			if max_n_features:
+				n_variables = min(n_variables, max_n_features)
+
+			if start_n_features is None:
+				perfs = [_ for _ in self.variable_selection_results['Running Achievable R-Squared'].astype(float)]
+				max_perf = np.max(perfs)
+				perf_threshold = start_n_features_perf_frac*max_perf
+				start_n_features = n_variables-len([_ for _ in perfs if _ > perf_threshold])+1
+
+			# 2. Sequentially add variables in decreasing order of importance.
+			# 2.1 Baseline performance
+			y_train = self.train_df[[self.target_column]].values.copy()
+			y_val = self.val_df[[self.target_column]].values.copy()
+			if return_scores:
+				y_test = self.test_df[[self.target_column]].values.copy()
+			x_train = self.train_df[self.variables[:1]].values.copy()
+			x_val = self.val_df[self.variables[:1]].values.copy()
+
+			base_m = BaselineRegressor(baseline=regression_baseline) if problem_type == 'regression' else BaselineClassifier()
+			base_m.fit(x_train, y_train)
+			y_val_pred = base_m.predict(x_val)
+			previous_score = score_func(y_val, y_val_pred)
+
+			spinner = Halo(text='Lean Boosting:', spinner='dots')
+			spinner.start()
+			logging.info('Baseline score (%s): %.4f' % (score, previous_score))
+			spinner.text = 'Lean Boosting -- Baseline %s: %.4f' % (score, previous_score)
+
+			self.start_n_features = min(start_n_features, n_variables)
+			n_down_perf = 0
+			for i in range(self.start_n_features, n_variables+1):
+				vs = self.variables[:i]
+				x_train = self.train_df[vs].values.copy()
+				x_val = self.val_df[vs].values.copy()
+				n_vars = x_train.shape[1] if len(x_train.shape) > 1 else 1
+
+				# Create the new model
+				m = learner_func(n_vars=n_vars)
+
+				# Fit the new model
+				m.fit(x_train, y_train)
+
+				# New validation score
+				y_val_pred = m.predict(x_val)
+				val_score = score_func(y_val, y_val_pred)
+
+				if val_score > previous_score or (min_n_features and i<=min_n_features):
+					n_down_perf = 0
+					logging.info('Variable #%d (%s) increased validation performance from %.4f to %.4f' % (i, self.variables[i-1], previous_score, val_score))
+					spinner.text = 'Lean Boosting -- %d Variables, Validation %s: %.4f' % (i, score, val_score)
+					previous_score = val_score
+					self.models = [m]
+					self.max_var_ixs = [i]
+					self.val_scores = self.val_scores + [(i, val_score)]
+
+				else:
+					n_down_perf += 1
+					logging.info('Validation performance did not increase for the %d-th consecutive time. Old: %.4f, New: %.4f, Variable: %s' % (n_down_perf, previous_score, val_score, self.variables[i-1]))
+					if n_down_perf >= n_down_perf_before_stop:
+						# Only stop after a certain number of consecutive down performance
+						logging.info('Stopping training as validation performance did not increase %d consecutive times.' % n_down_perf_before_stop)
+						spinner.succeed()
+						break
+
+				if max_n_features and (i==max_n_features):
+					logging.info('Stopping training as the maximum number of variables (%d) has been reached' % max_n_features)
+					spinner.succeed()
+					break		
+
+			self.selected_variables = self.variables[:self.max_var_ixs[-1]] if self.models else []
+			if self.models == []:
+				self.models = [base_m]
+				self.max_var_ixs = [1]
+				self.val_scores = [(0, previous_score)]
+
+			results = {'Selected Variables': self.selected_variables}
+			if return_scores:
+				# Inputs
+				x_train = self.train_df[self.selected_variables].values.copy()
+				x_val = self.val_df[self.selected_variables].values.copy()
+				x_test = self.test_df[self.selected_variables].values.copy()
+
+				# Predictions
+				self.y_train_pred = self.models[0].predict(x_train)
+				self.y_val_pred = self.models[0].predict(x_val)
+				self.y_test_pred = self.models[0].predict(x_test)
+
+				# Scores
+				self.train_score = score_func(y_train, self.y_train_pred)
+				self.val_score = score_func(y_val, self.y_val_pred)
+				self.test_score = score_func(y_test, self.y_test_pred)
+
+				results['Training Score'] = '%.5f' % self.train_score
+				results['Validation Score'] = '%.5f' % self.val_score
+				results['Testing Score'] = '%.5f' % self.test_score
+
+				if self.problem_type == 'regression':
+					results['Training R-Squared'] = '%.3f' % r2_score(y_train.flatten(), self.y_train_pred.flatten())
+					results['Validation R-Squared'] = '%.3f' % r2_score(y_val.flatten(), self.y_val_pred.flatten())
+					results['Testing R-Squared'] = '%.3f' % r2_score(y_test.flatten(), self.y_test_pred.flatten())
+
+					results['Training RMSE'] = '%.5f' % mean_squared_error(y_train.flatten(), self.y_train_pred.flatten(), squared=False)
+					results['Validation RMSE'] = '%.5f' % mean_squared_error(y_val.flatten(), self.y_val_pred.flatten(), squared=False)
+					results['Testing RMSE'] = '%.5f' % mean_squared_error(y_test.flatten(), self.y_test_pred.flatten(), squared=False)
+
+				if self.problem_type == 'classification':
+					results['Training Accuracy'] = '%.3f' % self.train_score
+					results['Validation Accuracy'] =  '%.3f' % self.val_score
+					results['Testing Accuracy'] = '%.3f' % self.test_score
+
+			return results
 
 
-	def _additive_fit(self):
-		pass
+
+
+	def _additive_fit(self, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
+			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=None, anonymize=False, \
+			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=1, \
+			regression_baseline='mean', regression_error_type='additive', return_scores=False, start_n_features_perf_frac=0.9):
+		# A base learner here is fitted to the residuals of the best model so far.
+		obj = self._obj
+		assert inspect.isfunction(learner_func), 'learner_func should be a class'
+		assert target_column in obj.columns, 'The target column should be a valid column'
+		if problem_type is None:
+			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'
+		assert problem_type in ('classification', 'regression')
+		self.problem_type = problem_type
+		self.additive_learning = True
+		assert regression_error_type in ('additive', 'multiplicative')
+		self.regression_error_type = regression_error_type
+
+		for col in obj.columns:
+			assert not self.is_categorical(col), 'All columns should be numeric'
+
+		x_columns = [_ for _ in obj.columns if _ != target_column]
+		if self.problem_type == 'classification':
+			labels = set(list(obj[target_column].values.astype(int)))
+			binary_labels = {0, 1}
+			assert labels.issubset(binary_labels), 'Classification labels should either be 0 or 1'
+
+		if benchmark_feature:
+			assert benchmark_feature in obj.columns, 'The benchmark feature should be a valid column'
+		self.benchmark_feature = benchmark_feature
+		if score == 'auto':
+			score = 'r2_score' if problem_type == 'regression' else 'accuracy_score'
+		score_func = eval(score)
+		self.val_scores = []
+
+		if getattr(self, 'models', None) is None or force_redo:
+			# 0. Train/Validation split
+			self.target_column = target_column
+			if return_scores:
+				# Reserve 1-train_frac for testing, and train_frac for training and validation
+				self.test_df = obj.sample(frac=1.-train_frac, random_state=random_state)
+				self.train_val_df = obj.drop(self.test_df.index)
+			else:
+				self.train_val_df = obj
+			# Reserve train_frac*train_frac for training [...]
+			self.train_df = self.train_val_df.sample(frac=train_frac, random_state=random_state)
+			# [...] and train_frac*(1-train_frac) for validation.
+			self.val_df = self.train_val_df.drop(self.train_df.index)
+
+			if missing_value_imputation:
+				# Basic missing value imputation
+				self.train_df.fillna(self.train_df.median(), inplace=True)
+				self.val_df.fillna(self.train_df.median(), inplace=True)
+				if return_scores:
+					self.test_df.fillna(self.train_df.median(), inplace=True)
+
+			# 1. Model-free variable selection
+			vs_accessor = PreLearningAccessor(obj)
+			self.variable_selection_results = vs_accessor.variable_selection(self.target_column, problem_type=self.problem_type, \
+				snr=snr, anonymize=anonymize)
+
+			self.variables = [_ for _ in self.variable_selection_results['Variable'].values if _.lower() != 'no variable']
+			n_variables = len(self.variables)
+			if max_n_features:
+				n_variables = min(n_variables, max_n_features)
+
+			if start_n_features is None:
+				perfs = [_ for _ in self.variable_selection_results['Running Achievable R-Squared'].astype(float)]
+				max_perf = np.max(perfs)
+				perf_threshold = start_n_features_perf_frac*max_perf
+				start_n_features = n_variables-len([_ for _ in perfs if _ > perf_threshold])+1
+
+			# 2. Sequentially add variables in decreasing order of importance.
+			# 2.1 Baseline performance
+			y_train = self.train_df[[self.target_column]].values.copy()
+			y_val = self.val_df[[self.target_column]].values.copy()
+			if return_scores:
+				y_test = self.test_df[[self.target_column]].values.copy()
+			x_train = self.train_df[self.variables[:1]].values.copy()
+			x_val = self.val_df[self.variables[:1]].values.copy()
+
+			base_m = BaselineRegressor(baseline=regression_baseline) if problem_type == 'regression' else BaselineClassifier()
+			base_m.fit(x_train, y_train)
+			y_val_pred = base_m.predict(x_val)
+			previous_score = score_func(y_val, y_val_pred)
+
+			spinner = Halo(text='Lean Boosting:', spinner='dots')
+			spinner.start()
+			logging.info('Baseline score (%s): %.4f' % (score, previous_score))
+			spinner.text = 'Lean Boosting -- Baseline %s: %.4f' % (score, previous_score)
+
+			self.start_n_features = min(start_n_features, n_variables)
+			n_down_perf = 0
+			target_train = y_train.copy()
+			y_val_pred = None
+			self.models = []
+			self.max_var_ixs = []
+			for i in range(self.start_n_features, n_variables+1):
+				vs = self.variables[:i]
+				x_train = self.train_df[vs].values.copy()
+				x_val = self.val_df[vs].values.copy()
+				n_vars = x_train.shape[1] if len(x_train.shape) > 1 else 1
+
+				# Create the new model
+				m = learner_func(n_vars=n_vars)
+
+				# Fit the new model
+				m.fit(x_train, target_train)
+
+				# New validation score
+				target_val_pred = m.predict(x_val)
+				target_val_pred = target_val_pred if len(target_val_pred.shape) > 1 else target_val_pred[:, None]
+				target_train_pred = m.predict(x_train)
+				target_train_pred = target_train_pred if len(target_train_pred.shape) > 1 else target_train_pred[:, None]
+
+				if y_val_pred is None:
+					y_val_pred = target_val_pred
+				else:
+					if self.problem_type == 'regression':
+						y_val_pred = y_val_pred+target_val_pred if self.regression_error_type == 'additive' else \
+							y_val_pred*target_val_pred
+
+					if self.problem_type == 'classification':
+						y_val_pred = np.abs(y_val_pred-target_val_pred)
+
+				val_score = score_func(y_val, y_val_pred)
+				if val_score > previous_score or (min_n_features and i<=min_n_features):
+					n_down_perf = 0
+					logging.info('Variable #%d (%s) increased validation performance from %.4f to %.4f' % (i, self.variables[i-1], previous_score, val_score))
+					spinner.text = 'Lean Boosting -- %d Variables, Validation %s: %.4f' % (i, score, val_score)
+					previous_score = val_score
+
+					if self.problem_type == 'regression':
+						target_train = target_train-target_train_pred if self.regression_error_type == 'additive' else target_train/target_train_pred
+
+					if self.problem_type == 'classification':
+						target_train = np.logical_not(target_train == target_train_pred).astype(int)
+
+					self.models = self.models+[m]
+					self.max_var_ixs = self.max_var_ixs+[i]
+					self.val_scores = self.val_scores + [(i, val_score)]
+
+				else:
+					n_down_perf += 1
+					logging.info('Validation performance did not increase for the %d-th consecutive time. Old: %.4f, New: %.4f, Variable: %s' % (n_down_perf, previous_score, val_score, self.variables[i-1]))
+					if n_down_perf >= n_down_perf_before_stop:
+						# Only stop after a certain number of consecutive down performance
+						logging.info('Stopping training as validation performance did not increase %d consecutive times.' % n_down_perf_before_stop)
+						spinner.succeed()
+						break
+
+				if max_n_features and (i==max_n_features):
+					logging.info('Stopping training as the maximum number of variables (%d) has been reached' % max_n_features)
+					spinner.succeed()
+					break	
+
+			self.selected_variables = self.variables[:self.max_var_ixs[-1]] if self.models else []
+			if self.models == []:
+				self.models = [base_m]
+				self.max_var_ixs = [1]
+				self.val_scores = [(0, previous_score)]
+
+			results = {'Selected Variables': self.selected_variables}
+			if return_scores:
+				# Inputs
+				x_train = self.train_df[self.selected_variables].values.copy()
+				x_val = self.val_df[self.selected_variables].values.copy()
+				x_test = self.test_df[self.selected_variables].values.copy()
+
+				# Predictions
+				self.y_train_pred = self.predict(self.train_df)
+				self.y_train_pred = self.y_train_pred.values.flatten()
+
+				self.y_val_pred = self.predict(self.val_df)
+				self.y_val_pred = self.y_val_pred.values.flatten()
+
+				self.y_test_pred = self.predict(self.test_df)
+				self.y_test_pred = self.y_test_pred.values.flatten()
+
+				# Scores
+				self.train_score = score_func(y_train, self.y_train_pred)
+				self.val_score = score_func(y_val, self.y_val_pred)
+				self.test_score = score_func(y_test, self.y_test_pred)
+
+				results['Training Score'] = '%.5f' % self.train_score
+				results['Validation Score'] = '%.5f' % self.val_score
+				results['Testing Score'] = '%.5f' % self.test_score
+
+				if self.problem_type == 'regression':
+					results['Training R-Squared'] = '%.3f' % r2_score(y_train.flatten(), self.y_train_pred.flatten())
+					results['Validation R-Squared'] = '%.3f' % r2_score(y_val.flatten(), self.y_val_pred.flatten())
+					results['Testing R-Squared'] = '%.3f' % r2_score(y_test.flatten(), self.y_test_pred.flatten())
+
+					results['Training RMSE'] = '%.5f' % mean_squared_error(y_train.flatten(), self.y_train_pred.flatten(), squared=False)
+					results['Validation RMSE'] = '%.5f' % mean_squared_error(y_val.flatten(), self.y_val_pred.flatten(), squared=False)
+					results['Testing RMSE'] = '%.5f' % mean_squared_error(y_test.flatten(), self.y_test_pred.flatten(), squared=False)
+
+				if self.problem_type == 'classification':
+					results['Training Accuracy'] = '%.3f' % self.train_score
+					results['Validation Accuracy'] =  '%.3f' % self.val_score
+					results['Testing Accuracy'] = '%.3f' % self.test_score
+
+			return results
+
 
 
 	def fit(self, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
-			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=1, anonymize=False, \
+			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=None, anonymize=False, \
 			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=1, \
-			regression_baseline='mean', additive_learning=False, regression_error_type='additive'):
+			regression_baseline='mean', additive_learning=False, regression_error_type='additive', return_scores=False, \
+			start_n_features_perf_frac=0.9):
 		"""
 		Train a lean boosted supervised learner, bringing in variables one at a time, in decreasing order of importance (as per :code:`df.kxy.variable_selection`), until doing so no longer improves validation performance or another stopping criterion is met.
 
@@ -114,10 +461,13 @@ class LearningAccessor(BaseAccessor):
 
 		Then we train a model (instance returned by :code:`learner_func`) using the :code:`start_n_features` most important feature/variable to predict the target (defined by :code:`target_column`).
 
-		Then we consider using the second most important variable to fix the mistakes made by the previously trained model.
+		Next we consider adding one variable at a time to fix the mistakes made by the previously trained model when :code:`additive_learning` is :code:`True`.
 	
 		If doing so improves performance on the validation set, we keep going until either performance no longer improves on the validation set :code:`n_down_perf_before_stop` consecutive times, or we've selected :code:`max_n_features` features.
 
+		When :code:`start_n_features` is :code:`None` the initial set of variables is the smallest set of variables with which we may achieve :code:`start_n_features_perf_frac` of the performance we could achieve using all variables (as per :code:`df.kxy.variable_selection`).
+
+		When :code:`additive_learning` is set to :code:`False`, after adding a new variable, we will train the new model on the original problem, rather than trying to improve residuals.
 
 
 		Parameters
@@ -157,6 +507,12 @@ class LearningAccessor(BaseAccessor):
 			When a new variable is added, whether errors/residuals should be fixed or a new model should be learned from scratch.
 		regression_error_type : str ('additive' | 'multiplicative')
 			For regression problems with additive learning, this determines whether the final model should be additive (pruning tries to reduce regressor residuals) or multiplicative (i.e. pruning tries to bring the ratio between true and predicted labels as closed to 1 as possible).
+		start_n_features_perf_frac : float (between 0 and 1)
+			When :code:`start_n_features` is not specified, it is set to the number of variables required to achieve a fraction :code:`start_n_features_perf_frac` of the maximum performance achievable (as per :code:`df.kxy.variable_selection`).
+		return_scores : bool (Default False)
+			Whether to return training, validation and testing performance after lean boosting.
+
+
 
 
 		Returns
@@ -164,231 +520,73 @@ class LearningAccessor(BaseAccessor):
 		result : dict
 			Dictionary containing selected variables, as well as training, validation and testing performance.
 		"""
-		obj = self._obj
-		assert inspect.isfunction(learner_func), 'learner_func should be a class'
-		assert target_column in obj.columns, 'The target column should be a valid column'
-		if problem_type is None:
-			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'
-		assert problem_type in ('classification', 'regression')
-		self.problem_type = problem_type
-		self.additive_learning = additive_learning
-		assert regression_error_type in ('additive', 'multiplicative')
-		self.regression_error_type = regression_error_type
 
+		if additive_learning:
+			return self._additive_fit(target_column, learner_func, problem_type=problem_type, snr=snr, train_frac=train_frac, random_state=random_state, \
+				force_redo=force_redo, max_n_features=max_n_features, min_n_features=min_n_features, start_n_features=start_n_features, anonymize=anonymize, \
+				benchmark_feature=benchmark_feature, missing_value_imputation=missing_value_imputation, score=score, n_down_perf_before_stop=n_down_perf_before_stop, \
+				regression_baseline=regression_baseline, regression_error_type=regression_error_type, return_scores=return_scores, start_n_features_perf_frac=start_n_features_perf_frac)
+
+		else:
+			return self._non_additive_fit(target_column, learner_func, problem_type=problem_type, snr=snr, train_frac=train_frac, random_state=random_state, \
+				force_redo=force_redo, max_n_features=max_n_features, min_n_features=min_n_features, start_n_features=start_n_features, anonymize=anonymize, \
+				benchmark_feature=benchmark_feature, missing_value_imputation=missing_value_imputation, score=score, n_down_perf_before_stop=n_down_perf_before_stop, \
+				regression_baseline=regression_baseline, regression_error_type=regression_error_type, return_scores=return_scores, start_n_features_perf_frac=start_n_features_perf_frac)
+
+
+
+	def _additive_predict(self, obj):
+		# Global predict in an additive setting (when base learners are fitted to residuals, not the original problem).
+		assert hasattr(self, 'models'), 'The model should be first fitted'
 		for col in obj.columns:
-			assert not self.is_categorical(col), 'All columns should be numeric'
+			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'
 
-		x_columns = [_ for _ in obj.columns if _ != target_column]
-		if self.problem_type == 'classification':
-			labels = set(list(obj[target_column].values.astype(int)))
-			binary_labels = {0, 1}
-			assert labels.issubset(binary_labels), 'Classification labels should either be 0 or 1'
+		data = obj[self.variables]
+		y_pred = None
+		for i in range(len(self.models)):
+			vs = self.variables[:self.max_var_ixs[i]]
+			x = data[vs].values.copy()
+			y_error_pred = self.models[i].predict(x)
+			y_error_pred = y_error_pred if len(y_error_pred.shape) > 1 else y_error_pred[:, None]
+			if y_pred is None:
+				y_pred = y_error_pred
+			else:
+				if self.problem_type == 'regression':
+					y_pred = y_pred+y_error_pred if self.regression_error_type == 'additive' else \
+						y_pred*y_error_pred
 
-		if benchmark_feature:
-			assert benchmark_feature in obj.columns, 'The benchmark feature should be a valid column'
-		self.benchmark_feature = benchmark_feature
-		if score == 'auto':
-			score = 'r2_score' if problem_type == 'regression' else 'accuracy_score'
-		score_func = eval(score)
+				if self.problem_type == 'classification':
+					y_pred = np.abs(y_pred-y_error_pred)
 
-		if getattr(self, 'models', None) is None or force_redo:
-			# 0. Train/Validation split
-			self.target_column = target_column
-			# Reserve 1-train_frac for testing, and train_frac for training and validation
-			self.test_df = obj.sample(frac=1.-train_frac, random_state=random_state)
-			self.train_val_df = obj.drop(self.test_df.index)
-			# Reserve train_frac*train_frac for training [...]
-			self.train_df = self.train_val_df.sample(frac=train_frac, random_state=random_state)
-			# [...] and train_frac*(1-train_frac) for validation.
-			self.val_df = self.train_val_df.drop(self.train_df.index)
+		y_pred = y_pred if len(y_pred.shape) > 1 else y_pred[:, None]
+		predictions = pd.DataFrame(y_pred, columns=[self.target_column], index=data.index)
 
-			if missing_value_imputation:
-				# Basic missing value imputation
-				self.train_df.fillna(self.train_df.median(), inplace=True)
-				self.val_df.fillna(self.train_df.median(), inplace=True)
-				self.test_df.fillna(self.train_df.median(), inplace=True)
-
-			# 1. Model-free variable selection
-			vs_accessor = PreLearningAccessor(obj)
-			self.variable_selection_results = vs_accessor.variable_selection(self.target_column, problem_type=self.problem_type, \
-				snr=snr, anonymize=anonymize)
-			self.variables = [_ for _ in self.variable_selection_results['Variable'].values if _.lower() != 'no variable']
-			n_variables = len(self.variables)
-			if max_n_features:
-				n_variables = min(n_variables, max_n_features)
-
-			# 2. Sequentially add variables in decreasing order of importance, each time with the aim to predict previous residuals.
-			y_train_df = self.train_df[[self.target_column]]
-			y_train = y_train_df.values.copy()
-			target_train = y_train.copy()
-			y_train_pred = None
-			y_val_df = self.val_df[[self.target_column]]
-			y_val = y_val_df.values.copy()
-			target_val = y_val.copy()
-			y_val_pred = None
-
-			models = []
-			max_var_ixs = []
-			m = BaselineRegressor(baseline=regression_baseline) if problem_type == 'regression' else BaselineClassifier()
-			x_train = self.train_df[self.variables[:1]].values.copy()
-			x_val = self.val_df[self.variables[:1]].values.copy()
-			m.fit(x_train, y_train)
-			target_train_pred = m.predict(x_train)
-			target_train_pred = target_train_pred if len(target_train_pred.shape) > 1 else target_train_pred[:, None]
-			y_train_pred = target_train_pred.copy()
-			target_val_pred = m.predict(x_val)
-			target_val_pred = target_val_pred if len(target_val_pred.shape) > 1 else target_val_pred[:, None]
-			y_val_pred = target_val_pred.copy()
-			previous_score = score_func(y_val, target_val_pred)
-			spinner = Halo(text='Lean Boosting:', spinner='dots')
-			spinner.start()
-			logging.info('Baseline score (%s): %.4f' % (score, previous_score))
-			spinner.text = 'Lean Boosting -- Baseline %s: %.4f' % (score, previous_score)
-
-			target_train = self._update_target(target_train, target_train_pred)
-			target_val = self._update_target(target_val, target_val_pred)
-			models = models+[m] if self.additive_learning else [m]
-			max_var_ixs = max_var_ixs+[0] if self.additive_learning else [0]	
-
-			self.start_n_features = start_n_features
-			n_down_perf = 0
-			for i in range(self.start_n_features, n_variables+1):
-				vs = self.variables[:i]
-				x_train = self.train_df[vs].values.copy()
-				x_val = self.val_df[vs].values.copy()
-				n_vars = x_train.shape[1] if len(x_train.shape) > 1 else 1
-
-				# Create the new model
-				m = learner_func(n_vars=n_vars)
-
-				# Fit the new model
-				m.fit(x_train, target_train)
-
-				# Error predictions
-				target_train_pred = m.predict(x_train)
-				target_train_pred = target_train_pred if len(target_train_pred.shape) > 1 else target_train_pred[:, None]
-				target_val_pred = m.predict(x_val)
-				target_val_pred = target_val_pred if len(target_val_pred.shape) > 1 else target_val_pred[:, None]
-
-				# Target predictions updates
-				y_train_pred_ = self._compound_predictions(y_train_pred, target_train_pred)
-				y_val_pred_ = self._compound_predictions(y_val_pred, target_val_pred)
-
-				# New validation score
-				val_score = score_func(y_val, y_val_pred_)
-				if val_score > previous_score or (min_n_features and i<=min_n_features):
-					y_train_pred = y_train_pred_.copy()
-					y_val_pred = y_val_pred_.copy()
-					n_down_perf = 0
-					logging.info('Variable #%d (%s) increased validation performance from %.4f to %.4f' % (i, self.variables[i-1], previous_score, val_score))
-					spinner.text = 'Lean Boosting -- %d Variables, Val %s: %.4f' % (i, score, val_score)
-					previous_score = val_score
-
-					target_train = self._update_target(target_train, target_train_pred)
-					target_val = self._update_target(target_val, target_val_pred)
-					models = models+[m] if self.additive_learning else [m]
-					max_var_ixs = max_var_ixs+[i] if self.additive_learning else [i]					
-
-				else:
-					n_down_perf += 1
-					logging.info('Validation performance did not increase for the %d-th consecutive time. Old: %.4f, New: %.4f, Variable: %s' % (n_down_perf, previous_score, val_score, self.variables[i-1]))
-					if n_down_perf >= n_down_perf_before_stop:
-						# Only stop after a certain number of consecutive down performance
-						logging.info('Stopping training as validation performance did not increase %d consecutive times.' % n_down_perf_before_stop)
-						spinner.succeed()
-						break
-
-				if max_n_features and (i==max_n_features):
-					logging.info('Stopping training as the maximum number of variables (%d) has been reached' % max_n_features)
-					spinner.succeed()
-					break				
-
-			self.models = models
-			self.max_var_ixs = max_var_ixs
-			self.selected_variables = self.variables[:self.max_var_ixs[-1]]
-
-			# Compute training/validation/testing performances
-			x_train_df = self.train_df[x_columns]
-			y_train_df = self.train_df[[target_column]]
-			train_predictions = self.predict(x_train_df)
-			self.train_score = score_func(y_train_df.values, train_predictions.values)
-			if self.problem_type == 'regression':
-				self.train_rmse = mean_squared_error(y_train_df.values.flatten(), train_predictions.values.flatten(), squared=False)
-				self.train_rmspe = rmspe_score(y_train_df.values.flatten(), train_predictions.values.flatten())
-				self.train_r2 = r2_score(y_train_df.values.flatten(), train_predictions.values.flatten())
-
-				if self.benchmark_feature:
-					train_benchmark = self.train_df[self.benchmark_feature]
-					self.train_benchmark_score = score_func(y_train_df.values, train_benchmark.values)
-					self.train_benchmark_rmse = mean_squared_error(y_train_df.values, train_benchmark.values, squared=False)
-					self.train_benchmark_rmspe = rmspe_score(y_train_df.values, train_benchmark.values)
-					self.train_benchmark_r2 = r2_score(y_train_df.values.flatten(), train_benchmark.values.flatten())
-
-			x_val_df = self.val_df[x_columns]
-			y_val_df = self.val_df[[target_column]]
-			val_predictions = self.predict(x_val_df)
-			self.val_score = score_func(y_train_df.values, train_predictions.values)
-			if self.problem_type == 'regression':
-				self.val_rmse = mean_squared_error(y_val_df.values.flatten(), val_predictions.values.flatten(), squared=False)
-				self.val_rmspe = rmspe_score(y_val_df.values.flatten(), val_predictions.values.flatten())
-				self.val_r2 = r2_score(y_val_df.values.flatten(), val_predictions.values.flatten())
-
-				if self.benchmark_feature:
-					val_benchmark = self.val_df[self.benchmark_feature]
-					self.val_benchmark_score = score_func(y_val_df.values, val_benchmark.values)
-					self.val_benchmark_rmse = mean_squared_error(y_val_df.values, val_benchmark.values, squared=False)
-					self.val_benchmark_rmspe = rmspe_score(y_val_df.values, val_benchmark.values)
-					self.val_benchmark_r2 = r2_score(y_val_df.values.flatten(), val_benchmark.values.flatten())
-				
-			x_test_df = self.test_df[x_columns]
-			y_test_df = self.test_df[[target_column]]
-			test_predictions = self.predict(x_test_df)
-			self.test_score = score_func(y_test_df.values, test_predictions.values)
-			if self.problem_type == 'regression':
-				self.test_rmse = mean_squared_error(y_test_df.values.flatten(), test_predictions.values.flatten(), squared=False)
-				self.test_rmspe = rmspe_score(y_test_df.values.flatten(), test_predictions.values.flatten())
-				self.test_r2 = r2_score(y_test_df.values.flatten(), test_predictions.values.flatten())
-
-				if self.benchmark_feature:
-					test_benchmark = self.test_df[self.benchmark_feature]
-					self.test_benchmark_score = score_func(y_test_df.values, test_benchmark.values)
-					self.test_benchmark_rmse = mean_squared_error(y_test_df.values, test_benchmark.values, squared=False)
-					self.test_benchmark_rmspe = rmspe_score(y_test_df.values, test_benchmark.values)
-					self.test_benchmark_r2 = r2_score(y_test_df.values.flatten(), test_benchmark.values.flatten())
+		return predictions
 
 
-		results = {'Selected Variables': self.selected_variables}
-		if self.problem_type == 'regression':
-			results['Training R-Squared'] = '%.3f' % self.train_r2
-			results['Validation R-Squared'] = '%.3f' % self.val_r2
-			results['Testing R-Squared'] = '%.3f' % self.test_r2
-			results['Training RMSE'] = '%.5f' % self.train_rmse
-			results['Validation RMSE'] = '%.5f' % self.val_rmse
-			results['Testing RMSE'] = '%.5f' % self.test_rmse
-			results['Training RMSPE'] = '%.5f' % self.train_rmspe
-			results['Validation RMSPE'] = '%.5f' % self.val_rmspe
-			results['Testing RMSPE'] = '%.5f' % self.test_rmspe
-			if self.benchmark_feature:
-				results['Benchmark Training R-Squared'] = '%.3f' % self.train_benchmark_r2
-				results['Benchmark Validation R-Squared'] = '%.3f' % self.val_benchmark_r2
-				results['Benchmark Testing R-Squared'] = '%.3f' % self.test_benchmark_r2
-				results['Benchmark Training RMSE'] = '%.5f' % self.train_benchmark_rmse
-				results['Benchmark Validation RMSE'] = '%.5f' % self.val_benchmark_rmse
-				results['Benchmark Testing RMSE'] = '%.5f' % self.test_benchmark_rmse
-				results['Benchmark Training RMSPE'] = '%.5f' % self.train_benchmark_rmspe
-				results['Benchmark Validation RMSPE'] = '%.5f' % self.val_benchmark_rmspe
-				results['Benchmark Testing RMSPE'] = '%.5f' % self.test_benchmark_rmspe
+	def _non_additive_predict(self, obj):
+		# Global predict in a non-additive setting (when base learners are fitted to the original problem, no residuals).
+		assert hasattr(self, 'models'), 'The model should be first fitted'
+		for col in obj.columns:
+			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'
 
-		if self.problem_type == 'classification':
-			results['Training Accuracy'] = '%.3f' % self.train_score
-			results['Validation Accuracy'] = '%.3f' % self.val_score
-			results['Testing Accuracy'] = '%.3f' % self.test_score
+		data = obj[self.variables]
+		x = data[self.selected_variables].values.copy()
+		y_pred = self.models[0].predict(x)
+		y_pred = y_pred if len(y_pred.shape) > 1 else y_pred[:, None]
+		predictions = pd.DataFrame(y_pred, columns=[self.target_column], index=data.index)
 
-		return results
-		
+		return predictions
 
 
-	def predict(self, obj):
+	def _predict(self, obj):
+		if self.additive_learning:
+			return self._additive_predict(obj)
+		else:
+			return self._non_additive_predict(obj)
+
+
+	def predict(self, obj, memory_bound=False):
 		"""
 		Make predictions using the fitted model.
 
@@ -397,6 +595,8 @@ class LearningAccessor(BaseAccessor):
 		----------
 		obj : pandas.DataFrame
 			A dataframe containing test explanatory variables about which we want to make predictions.
+		memory_bound : bool (Default False)
+			Whether we should try to save memory.
 
 
 		Returns
@@ -404,52 +604,19 @@ class LearningAccessor(BaseAccessor):
 		result : pandas.DataFrame
 			A dataframe with the same index as :code:`obj`, and with one column whose name is the :code:`target_column` used for training.
 		"""
-		assert hasattr(self, 'models'), 'The model should be first fitted'
-		for col in obj.columns:
-			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'
-
-		data = obj[self.variables]
-
-		if self.additive_learning:
-			x = data[self.variables[:1]].values.copy()
-			y_error_pred = self.models[0].predict(x)
-			y_error_pred = y_error_pred if len(y_error_pred.shape) > 1 else y_error_pred[:, None]
-			y_pred = y_error_pred
-
-			for i in range(1, len(self.models)):
-				vs = self.variables[:self.max_var_ixs[i]]
-				x = data[vs].values.copy()
-				y_error_pred = self.models[i].predict(x)
-				y_error_pred = y_error_pred if len(y_error_pred.shape) > 1 else y_error_pred[:, None]
-				y_pred = self._compound_predictions(y_pred, y_error_pred)
+		if memory_bound:
+			n = obj.shape[0]
+			max_n = 1000000
+			res = pd.DataFrame(index=obj.index)
+			res[self.target_column] = np.nan
+			i = 0
+			while i < n:
+				res.iloc[i:i+max_n] = self._predict(obj.iloc[i:i+max_n])
+				i += max_n
+			return res
 
 		else:
-			x = data[self.selected_variables].values.copy()
-			y_pred = self.models[0].predict(x)
-			y_pred = y_pred if len(y_pred.shape) > 1 else y_pred[:, None]
-
-		predictions = pd.DataFrame(y_pred, columns=[self.target_column], index=data.index)
-
-		return predictions
-
-
-
-	def batch_predict(self, obj):
-		"""
-		Same as :code:`predict(self, obj)`, but more memory friendly.
-		"""
-		n = obj.shape[0]
-		max_n = 100000
-		res = []
-		i = 0
-		while i < n:
-			res +=[self.predict(obj.iloc[i:i+max_n])]
-			i += max_n
-
-		res = pd.concat(res, axis=0) if len(res) > 1 else res[0]
-		return res
-
-
+			return self._predict(obj)
 
 
 
